@@ -50,14 +50,6 @@ or
 
 # run echo client to test IETF HyBi 00 protocol
  run with --protocol-version=hybi00
-
-or
-
-# server setup to test Hixie 75 protocol
- run with --allow-draft75
-
-# run echo client to test Hixie 75 protocol
- run with --protocol-version=hixie75
 """
 
 
@@ -95,6 +87,15 @@ _PROTOCOL_VERSION_HYBI08 = 'hybi08'
 _PROTOCOL_VERSION_HYBI00 = 'hybi00'
 _PROTOCOL_VERSION_HIXIE75 = 'hixie75'
 
+# Constants for the --tls_module flag.
+_TLS_BY_STANDARD_MODULE = 'ssl'
+_TLS_BY_PYOPENSSL = 'pyopenssl'
+
+# Values used by the --tls-version flag.
+_TLS_VERSION_SSL23 = 'ssl23'
+_TLS_VERSION_SSL3 = 'ssl3'
+_TLS_VERSION_TLS1 = 'tls1'
+
 
 class ClientHandshakeError(Exception):
     pass
@@ -183,18 +184,71 @@ def _validate_mandatory_header(fields, name,
 class _TLSSocket(object):
     """Wrapper for a TLS connection."""
 
-    def __init__(self, raw_socket):
-        self._ssl = socket.ssl(raw_socket)
+    def __init__(self,
+                 raw_socket, tls_module, tls_version, disable_tls_compression):
+        self._logger = util.get_class_logger(self)
 
-    def send(self, bytes):
-        return self._ssl.write(bytes)
+        if tls_module == _TLS_BY_STANDARD_MODULE:
+            if tls_version == _TLS_VERSION_SSL23:
+                version = ssl.PROTOCOL_SSLv23
+            elif tls_version == _TLS_VERSION_SSL3:
+                version = ssl.PROTOCOL_SSLv3
+            elif tls_version == _TLS_VERSION_TLS1:
+                version = ssl.PROTOCOL_TLSv1
+            else:
+                raise ValueError(
+                    'Invalid --tls-version flag: %r' % tls_version)
+
+            if disable_tls_compression:
+                raise ValueError(
+                    '--disable-tls-compression is not available for ssl '
+                    'module')
+
+            self._tls_socket = ssl.wrap_socket(raw_socket, ssl_version=version)
+
+            # Print cipher in use. Handshake is done on wrap_socket call.
+            self._logger.info("Cipher: %s", self._tls_socket.cipher())
+        elif tls_module == _TLS_BY_PYOPENSSL:
+            if tls_version == _TLS_VERSION_SSL23:
+                version = OpenSSL.SSL.SSLv23_METHOD
+            elif tls_version == _TLS_VERSION_SSL3:
+                version = OpenSSL.SSL.SSLv3_METHOD
+            elif tls_version == _TLS_VERSION_TLS1:
+                version = OpenSSL.SSL.TLSv1_METHOD
+            else:
+                raise ValueError(
+                    'Invalid --tls-version flag: %r' % tls_version)
+
+            context = OpenSSL.SSL.Context(version)
+
+            if disable_tls_compression:
+                # OP_NO_COMPRESSION is not defined in OpenSSL module.
+                context.set_options(0x00020000)
+
+            self._tls_socket = OpenSSL.SSL.Connection(context, raw_socket)
+            # Client mode.
+            self._tls_socket.set_connect_state()
+            self._tls_socket.setblocking(True)
+
+            # Do handshake now (not necessary).
+            self._tls_socket.do_handshake()
+        else:
+            raise ValueError('No TLS support module is available')
+
+    def send(self, data):
+        return self._tls_socket.write(data)
+
+    def sendall(self, data):
+        return self._tls_socket.sendall(data)
 
     def recv(self, size=-1):
-        return self._ssl.read(size)
+        return self._tls_socket.read(size)
 
     def close(self):
-        # Nothing to do.
-        pass
+        return self._tls_socket.close()
+
+    def getpeername(self):
+        return self._tls_socket.getpeername()
 
 
 class ClientHandshakeBase(object):
@@ -697,66 +751,6 @@ class ClientHandshakeProcessorHybi00(ClientHandshakeBase):
         return ''.join([chr(random.randint(0, 255)) for _ in xrange(8)])
 
 
-class ClientHandshakeProcessorHixie75(object):
-    """WebSocket opening handshake processor for
-    draft-hixie-thewebsocketprotocol-75.
-    """
-
-    _EXPECTED_RESPONSE = (
-        'HTTP/1.1 101 Web Socket Protocol Handshake\r\n' +
-        _UPGRADE_HEADER_HIXIE75 +
-        _CONNECTION_HEADER)
-
-    def __init__(self, socket, options):
-        self._socket = socket
-        self._options = options
-
-        self._logger = util.get_class_logger(self)
-
-    def _skip_headers(self):
-        terminator = '\r\n\r\n'
-        pos = 0
-        while pos < len(terminator):
-            received = _receive_bytes(self._socket, 1)
-            if received == terminator[pos]:
-                pos += 1
-            elif received == terminator[0]:
-                pos = 1
-            else:
-                pos = 0
-
-    def handshake(self):
-        """Performs opening handshake on the specified socket.
-
-        Raises:
-            ClientHandshakeError: handshake failed.
-        """
-
-        self._socket.sendall(_build_method_line(self._options.resource))
-        self._socket.sendall(_UPGRADE_HEADER_HIXIE75)
-        self._socket.sendall(_CONNECTION_HEADER)
-        self._socket.sendall(_format_host_header(
-            self._options.server_host,
-            self._options.server_port,
-            self._options.use_tls))
-        if not self._options.origin:
-            raise ClientHandshakeError(
-                'Specify the origin of the connection by --origin flag')
-        self._socket.sendall(_origin_header(common.ORIGIN_HEADER,
-                                            self._options.origin))
-        self._socket.sendall('\r\n')
-
-        self._logger.info('Sent handshake')
-
-        for expected_char in (
-            ClientHandshakeProcessorHixie75._EXPECTED_RESPONSE):
-            received = _receive_bytes(self._socket, 1)
-            if expected_char != received:
-                raise ClientHandshakeError('Handshake failure')
-        # We cut corners and skip other headers.
-        self._skip_headers()
-
-
 class ClientConnection(object):
     """A wrapper for socket object to provide the mp_conn interface.
     mod_pywebsocket library is designed to be working on Apache mod_python's
@@ -798,6 +792,24 @@ class ClientRequest(object):
                 'Drained data following close frame: %r', drained_data)
 
 
+def _import_ssl():
+    global ssl
+    try:
+        import ssl
+        return True
+    except ImportError:
+        return False
+
+
+def _import_pyopenssl():
+    global OpenSSL
+    try:
+        import OpenSSL.SSL
+        return True
+    except ImportError:
+        return False
+
+
 class EchoClient(object):
     """WebSocket echo client."""
 
@@ -819,7 +831,11 @@ class EchoClient(object):
             self._socket.connect((self._options.server_host,
                                   self._options.server_port))
             if self._options.use_tls:
-                self._socket = _TLSSocket(self._socket)
+                self._socket = _TLSSocket(
+                    self._socket,
+                    self._options.tls_module,
+                    self._options.tls_version,
+                    self._options.disable_tls_compression)
 
             version = self._options.protocol_version
 
@@ -830,9 +846,6 @@ class EchoClient(object):
             elif version == _PROTOCOL_VERSION_HYBI00:
                 self._handshake = ClientHandshakeProcessorHybi00(
                     self._socket, self._options)
-            elif version == _PROTOCOL_VERSION_HIXIE75:
-                self._handshake = ClientHandshakeProcessorHixie75(
-                    self._socket, self._options)
             else:
                 raise ValueError(
                     'Invalid --protocol-version flag: %r' % version)
@@ -846,8 +859,7 @@ class EchoClient(object):
             version_map = {
                 _PROTOCOL_VERSION_HYBI08: common.VERSION_HYBI08,
                 _PROTOCOL_VERSION_HYBI13: common.VERSION_HYBI13,
-                _PROTOCOL_VERSION_HYBI00: common.VERSION_HYBI00,
-                _PROTOCOL_VERSION_HIXIE75: common.VERSION_HIXIE75}
+                _PROTOCOL_VERSION_HYBI00: common.VERSION_HYBI00}
             request.ws_version = version_map[version]
 
             if (version == _PROTOCOL_VERSION_HYBI08 or
@@ -866,8 +878,6 @@ class EchoClient(object):
                 self._stream = Stream(request, stream_option)
             elif version == _PROTOCOL_VERSION_HYBI00:
                 self._stream = StreamHixie75(request, True)
-            elif version == _PROTOCOL_VERSION_HIXIE75:
-                self._stream = StreamHixie75(request)
 
             for line in self._options.message.split(','):
                 self._stream.send_message(line)
@@ -883,8 +893,7 @@ class EchoClient(object):
                         print 'Error: %s' % e
                     raise
 
-            if version != _PROTOCOL_VERSION_HIXIE75:
-                self._do_closing_handshake()
+            self._do_closing_handshake()
         finally:
             self._socket.close()
 
@@ -932,22 +941,41 @@ def main():
     parser.add_option('-q', '--quiet', dest='verbose', action='store_false',
                       default=True, help='suppress messages')
     parser.add_option('-t', '--tls', dest='use_tls', action='store_true',
-                      default=False, help='use TLS (wss://)')
+                      default=False, help='use TLS (wss://). By default, '
+                      'it looks for ssl and pyOpenSSL module and uses found '
+                      'one. Use --tls-module option to specify which module '
+                      'to use')
+    parser.add_option('--tls-module', '--tls_module', dest='tls_module',
+                      type='choice',
+                      choices = [_TLS_BY_STANDARD_MODULE, _TLS_BY_PYOPENSSL],
+                      help='Use ssl module if "%s" is specified. '
+                      'Use pyOpenSSL module if "%s" is specified' %
+                      (_TLS_BY_STANDARD_MODULE, _TLS_BY_PYOPENSSL))
+    parser.add_option('--tls-version', '--tls_version',
+                      dest='tls_version',
+                      type='string', default=_TLS_VERSION_SSL23,
+                      help='TLS/SSL version to use. One of \'' +
+                      _TLS_VERSION_SSL23 + '\' (SSL version 2 or 3), \'' +
+                      _TLS_VERSION_SSL3 + '\' (SSL version 3), \'' +
+                      _TLS_VERSION_TLS1 + '\' (TLS version 1)')
+    parser.add_option('--disable-tls-compression', '--disable_tls_compression',
+                      dest='disable_tls_compression',
+                      action='store_true', default=False,
+                      help='Disable TLS compression. Available only when '
+                      'pyOpenSSL module is used.')
     parser.add_option('-k', '--socket-timeout', '--socket_timeout',
                       dest='socket_timeout', type='int', default=_TIMEOUT_SEC,
                       help='Timeout(sec) for sockets')
     parser.add_option('--draft75', dest='draft75',
-                       action='store_true', default=False,
-                      help='use the Hixie 75 protocol. This overrides '
-                      'protocol-version flag')
+                      action='store_true', default=False,
+                      help='Obsolete option. Don\'t use this.')
     parser.add_option('--protocol-version', '--protocol_version',
                       dest='protocol_version',
                       type='string', default=_PROTOCOL_VERSION_HYBI13,
                       help='WebSocket protocol version to use. One of \'' +
                       _PROTOCOL_VERSION_HYBI13 + '\', \'' +
                       _PROTOCOL_VERSION_HYBI08 + '\', \'' +
-                      _PROTOCOL_VERSION_HYBI00 + '\', \'' +
-                      _PROTOCOL_VERSION_HIXIE75 + '\'')
+                      _PROTOCOL_VERSION_HYBI00 + '\'')
     parser.add_option('--version-header', '--version_header',
                       dest='version_header',
                       type='int', default=-1,
@@ -974,7 +1002,55 @@ def main():
     logging.basicConfig(level=logging.getLevelName(options.log_level.upper()))
 
     if options.draft75:
-        options.protocol_version = _PROTOCOL_VERSION_HIXIE75
+        logging.critical('--draft75 option is obsolete.')
+        sys.exit(1)
+
+    if options.protocol_version == _PROTOCOL_VERSION_HIXIE75:
+        logging.critical(
+            'Value %s is obsolete for --protocol_version options' %
+            _PROTOCOL_VERSION_HIXIE75)
+        sys.exit(1)
+
+    if options.use_tls:
+        if options.tls_module is None:
+            if _import_ssl():
+                options.tls_module = _TLS_BY_STANDARD_MODULE
+                logging.debug('Using ssl module')
+            elif _import_pyopenssl():
+                options.tls_module = _TLS_BY_PYOPENSSL
+                logging.debug('Using pyOpenSSL module')
+            else:
+                logging.critical(
+                        'TLS support requires ssl or pyOpenSSL module.')
+                sys.exit(1)
+        elif options.tls_module == _TLS_BY_STANDARD_MODULE:
+            if not _import_ssl():
+                logging.critical('ssl module is not available')
+                sys.exit(1)
+        elif options.tls_module == _TLS_BY_PYOPENSSL:
+            if not _import_pyopenssl():
+                logging.critical('pyOpenSSL module is not available')
+                sys.exit(1)
+        else:
+            logging.critical('Invalid --tls-module option: %r',
+                             options.tls_module)
+            sys.exit(1)
+
+        if (options.disable_tls_compression and
+            options.tls_module != _TLS_BY_PYOPENSSL):
+            logging.critical('You can disable TLS compression only when '
+                             'pyOpenSSL module is used.')
+            sys.exit(1)
+    else:
+        if options.tls_module is not None:
+            logging.critical('Use --tls-module option only together with '
+                             '--use-tls option.')
+            sys.exit(1)
+
+        if options.disable_tls_compression:
+            logging.critical('Use --disable-tls-compression only together '
+                             'with --use-tls option.')
+            sys.exit(1)
 
     # Default port number depends on whether TLS is used.
     if options.server_port == _UNDEFINED_PORT:
