@@ -45,6 +45,7 @@ import unittest
 import set_sys_path  # Update sys.path to locate mod_pywebsocket module.
 
 from test import client_for_testing
+from test import mux_client_for_testing
 
 
 # Special message that tells the echo server to start closing handshake
@@ -118,6 +119,40 @@ def _unmasked_frame_check_procedure(client):
     client.assert_connection_closed()
 
 
+def _mux_echo_check_procedure(mux_client):
+    mux_client.connect()
+    mux_client.send_flow_control(1, 1024)
+
+    logical_channel_options = client_for_testing.ClientOptions()
+    logical_channel_options.server_host = 'localhost'
+    logical_channel_options.server_port = 80
+    logical_channel_options.origin = 'http://localhost'
+    logical_channel_options.resource = '/echo'
+    mux_client.add_channel(2, logical_channel_options)
+    mux_client.send_flow_control(2, 1024)
+
+    mux_client.send_message(2, 'test')
+    mux_client.assert_receive(2, 'test')
+
+    mux_client.add_channel(3, logical_channel_options)
+    mux_client.send_flow_control(3, 1024)
+
+    mux_client.send_message(2, 'hello')
+    mux_client.send_message(3, 'world')
+    mux_client.assert_receive(2, 'hello')
+    mux_client.assert_receive(3, 'world')
+
+    # Don't send close message on channel id 1 so that server-initiated
+    # closing handshake won't occur.
+    mux_client.send_close(2)
+    mux_client.send_close(3)
+    mux_client.assert_receive_close(2)
+    mux_client.assert_receive_close(3)
+
+    mux_client.send_physical_connection_close()
+    mux_client.assert_physical_connection_receive_close()
+
+
 class EndToEndTest(unittest.TestCase):
     """An end-to-end test that launches pywebsocket standalone server as a
     separate process, connects to it using the client_for_testing module, and
@@ -205,7 +240,7 @@ class EndToEndTest(unittest.TestCase):
         try:
             time.sleep(0.2)
 
-            self._options.use_deflate_stream = True
+            self._options.enable_deflate_stream()
             client = client_for_testing.create_client(self._options)
             try:
                 test_function(client)
@@ -219,7 +254,7 @@ class EndToEndTest(unittest.TestCase):
         try:
             time.sleep(0.2)
 
-            self._options.use_deflate_frame = True
+            self._options.enable_deflate_frame()
             client = client_for_testing.create_client(self._options)
             try:
                 test_function(client)
@@ -260,6 +295,19 @@ class EndToEndTest(unittest.TestCase):
         finally:
             self._kill_process(server.pid)
 
+    def _run_hybi_mux_test(self, test_function):
+        server = self._run_server()
+        try:
+            time.sleep(0.2)
+
+            client = mux_client_for_testing.MuxClient(self._options)
+            try:
+                test_function(client)
+            finally:
+                client.close_socket()
+        finally:
+            self._kill_process(server.pid)
+
     def test_echo(self):
         self._run_hybi_test(_echo_check_procedure)
 
@@ -295,6 +343,9 @@ class EndToEndTest(unittest.TestCase):
         self._run_hybi_close_with_code_and_reason_test(
             _echo_check_procedure_with_code_and_reason, None, '')
 
+    def test_mux_echo(self):
+        self._run_hybi_mux_test(_mux_echo_check_procedure)
+
     def test_close_on_protocol_error(self):
         """Tests that the server sends a close frame with protocol error status
         code when the client sends data with some protocol error.
@@ -343,6 +394,20 @@ class EndToEndTest(unittest.TestCase):
 
         self._run_hybi_test(test_function)
 
+    def test_close_on_internal_endpoint_error(self):
+        """Tests that the server sends a close frame with internal endpoint
+        error status code when the handler does bad operation.
+        """
+
+        self._options.resource = '/internal_error'
+
+        def test_function(client):
+            client.connect()
+            client.assert_receive_close(
+                client_for_testing.STATUS_INTERNAL_ENDPOINT_ERROR)
+
+        self._run_hybi_test(test_function)
+
     def _run_hybi00_test(self, test_function):
         server = self._run_server()
         try:
@@ -362,50 +427,6 @@ class EndToEndTest(unittest.TestCase):
     def test_echo_server_close_hybi00(self):
         self._run_hybi00_test(_echo_check_procedure_with_goodbye)
 
-    def _run_hixie75_test(self, test_function):
-        server = self._run_server(allow_draft75=True)
-        try:
-            time.sleep(0.2)
-
-            client = client_for_testing.create_client_hixie75(self._options)
-            try:
-                test_function(client)
-            finally:
-                client.close_socket()
-        finally:
-            self._kill_process(server.pid)
-
-    def test_echo_hixie75(self):
-        """Tests that the server can talk draft-hixie-thewebsocketprotocol-75
-        protocol.
-        """
-
-        def test_function(client):
-            client.connect()
-
-            client.send_message('test')
-            client.assert_receive('test')
-
-        self._run_hixie75_test(test_function)
-
-    def test_echo_server_close_hixie75(self):
-        """Tests that the server can talk draft-hixie-thewebsocketprotocol-75
-        protocol. At the end of message exchanging, the client sends a keyword
-        message that requests the server to close the connection, and then
-        checks if the connection is really closed.
-        """
-
-        def test_function(client):
-            client.connect()
-
-            client.send_message('test')
-            client.assert_receive('test')
-
-            client.send_message(_GOODBYE_MESSAGE)
-            client.assert_receive(_GOODBYE_MESSAGE)
-
-        self._run_hixie75_test(test_function)
-
     # TODO(toyoshim): Add tests to verify invalid absolute uri handling like
     # host unmatch, port unmatch and invalid port description (':' without port
     # number).
