@@ -159,6 +159,18 @@ class HandshakerTest(unittest.TestCase):
         self.assertEqual(None, request.ws_extensions)
         self.assertEqual(common.VERSION_HYBI_LATEST, request.ws_version)
 
+    def test_do_handshake_with_extra_headers(self):
+        request_def = _create_good_request_def()
+        # Add headers not related to WebSocket opening handshake.
+        request_def.headers['FooKey'] = 'BarValue'
+        request_def.headers['EmptyKey'] = ''
+
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+        self.assertEqual(
+            _EXPECTED_RESPONSE, request.connection.written_data())
+
     def test_do_handshake_with_capitalized_value(self):
         request_def = _create_good_request_def()
         request_def.headers['upgrade'] = 'WEBSOCKET'
@@ -250,8 +262,39 @@ class HandshakerTest(unittest.TestCase):
         self.assertEqual(EXPECTED_RESPONSE, request.connection.written_data())
         self.assertEqual(1, len(request.ws_extensions))
         extension = request.ws_extensions[0]
-        self.assertEqual('deflate-stream', extension.name())
+        self.assertEqual(common.DEFLATE_STREAM_EXTENSION, extension.name())
         self.assertEqual(0, len(extension.get_parameter_names()))
+        self.assertEqual(1, len(request.ws_extension_processors))
+        self.assertEqual(common.DEFLATE_STREAM_EXTENSION,
+                         request.ws_extension_processors[0].name())
+
+    def test_do_handshake_with_perframe_compress(self):
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Extensions'] = (
+            'perframe-compress; method=deflate')
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+        self.assertEqual(1, len(request.ws_extensions))
+        self.assertEqual(common.PERFRAME_COMPRESSION_EXTENSION,
+                         request.ws_extensions[0].name())
+        self.assertEqual(1, len(request.ws_extension_processors))
+        self.assertEqual(common.PERFRAME_COMPRESSION_EXTENSION,
+                         request.ws_extension_processors[0].name())
+
+    def test_do_handshake_with_permessage_compress(self):
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Extensions'] = (
+            'permessage-compress; method=deflate')
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+        self.assertEqual(1, len(request.ws_extensions))
+        self.assertEqual(common.PERMESSAGE_COMPRESSION_EXTENSION,
+                         request.ws_extensions[0].name())
+        self.assertEqual(1, len(request.ws_extension_processors))
+        self.assertEqual(common.PERMESSAGE_COMPRESSION_EXTENSION,
+                         request.ws_extension_processors[0].name())
 
     def test_do_handshake_with_quoted_extensions(self):
         request_def = _create_good_request_def()
@@ -295,6 +338,91 @@ class HandshakerTest(unittest.TestCase):
         # not caught by do_handshake.
         self.assertRaises(AbortedByUserException, handshaker.do_handshake)
 
+    def test_do_handshake_with_mux_and_deflate_frame(self):
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Extensions'] = ('%s, %s' % (
+                common.MUX_EXTENSION,
+                common.DEFLATE_FRAME_EXTENSION))
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+        # mux should be rejected.
+        self.assertEqual(1, len(request.ws_extensions))
+        self.assertEqual(common.DEFLATE_FRAME_EXTENSION,
+                         request.ws_extensions[0].name())
+        self.assertEqual(2, len(request.ws_extension_processors))
+        self.assertEqual(common.MUX_EXTENSION,
+                         request.ws_extension_processors[0].name())
+        self.assertEqual(common.DEFLATE_FRAME_EXTENSION,
+                         request.ws_extension_processors[1].name())
+        self.assertFalse(hasattr(request, 'mux_processor'))
+
+    def test_do_handshake_with_deflate_frame_and_mux(self):
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Extensions'] = ('%s, %s' % (
+                common.DEFLATE_FRAME_EXTENSION,
+                common.MUX_EXTENSION))
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+        # mux should be rejected.
+        self.assertEqual(1, len(request.ws_extensions))
+        first_extension = request.ws_extensions[0]
+        self.assertEqual(common.DEFLATE_FRAME_EXTENSION,
+                         first_extension.name())
+        self.assertEqual(2, len(request.ws_extension_processors))
+        self.assertEqual(common.DEFLATE_FRAME_EXTENSION,
+                         request.ws_extension_processors[0].name())
+        self.assertEqual(common.MUX_EXTENSION,
+                         request.ws_extension_processors[1].name())
+        self.assertFalse(hasattr(request, 'mux'))
+
+    def test_do_handshake_with_permessage_compress_and_mux(self):
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Extensions'] = (
+            '%s; method=deflate, %s' % (
+                common.PERMESSAGE_COMPRESSION_EXTENSION,
+                common.MUX_EXTENSION))
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+
+        self.assertEqual(1, len(request.ws_extensions))
+        self.assertEqual(common.MUX_EXTENSION,
+                         request.ws_extensions[0].name())
+        self.assertEqual(2, len(request.ws_extension_processors))
+        self.assertEqual(common.PERMESSAGE_COMPRESSION_EXTENSION,
+                         request.ws_extension_processors[0].name())
+        self.assertEqual(common.MUX_EXTENSION,
+                         request.ws_extension_processors[1].name())
+        self.assertTrue(hasattr(request, 'mux_processor'))
+        self.assertTrue(request.mux_processor.is_active())
+        mux_extensions = request.mux_processor.extensions()
+        self.assertEqual(1, len(mux_extensions))
+        self.assertEqual(common.PERMESSAGE_COMPRESSION_EXTENSION,
+                         mux_extensions[0].name())
+
+    def test_do_handshake_with_mux_and_permessage_compress(self):
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Extensions'] = (
+            '%s, %s; method=deflate' % (
+                common.MUX_EXTENSION,
+                common.PERMESSAGE_COMPRESSION_EXTENSION))
+        request = _create_request(request_def)
+        handshaker = _create_handshaker(request)
+        handshaker.do_handshake()
+        # mux should be rejected.
+        self.assertEqual(1, len(request.ws_extensions))
+        first_extension = request.ws_extensions[0]
+        self.assertEqual(common.PERMESSAGE_COMPRESSION_EXTENSION,
+                         first_extension.name())
+        self.assertEqual(2, len(request.ws_extension_processors))
+        self.assertEqual(common.MUX_EXTENSION,
+                         request.ws_extension_processors[0].name())
+        self.assertEqual(common.PERMESSAGE_COMPRESSION_EXTENSION,
+                         request.ws_extension_processors[1].name())
+        self.assertFalse(hasattr(request, 'mux_processor'))
+
     def test_bad_requests(self):
         bad_cases = [
             ('HTTP request',
@@ -384,6 +512,16 @@ class HandshakerTest(unittest.TestCase):
         bad_cases.append(('Wrong Sec-WebSocket-Version (multiple values)',
                           request_def, 400, True))
 
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Protocol'] = 'illegal\x09protocol'
+        bad_cases.append(('Illegal Sec-WebSocket-Protocol',
+                          request_def, 400, True))
+
+        request_def = _create_good_request_def()
+        request_def.headers['Sec-WebSocket-Protocol'] = ''
+        bad_cases.append(('Empty Sec-WebSocket-Protocol',
+                          request_def, 400, True))
+
         for (case_name, request_def, expected_status,
              expect_handshake_exception) in bad_cases:
             request = _create_request(request_def)
